##################################################################################################
######################################## CLASS DEFINITION ########################################
##################################################################################################
Database <- setRefClass(
  Class = "Database",
  fields = list(
    data="data.frame",
    maxStatesDiscreteVariable = "numeric"
  )
)

##################################################################################################
###################################### SETTERS AND GETTERS #######################################
##################################################################################################
Database$accessors(c("data", "maxStatesDiscreteVariable"))

##################################################################################################
########################################### METHODS ##############################################
##################################################################################################

##################################################################################################

# Initializes the object.
#
# Args:
#   path: A String with the file path (non mandatory argument). Data columns are separated by ";"
#         and a header indicates the variables names.
#   data: An object of type "Database" (non mandatory argument).
#   maxStateDiscreteVar: The maximum number of states to be allowed in a discrete variable.  If
#                        the number of states in the data is greater, the data type is considered
#                        as continuous (non mandatory argument).
#
# Returns:
#   The data frame.

Database$methods (
  initialize = function(path=NULL, data=NULL, maxStatesDiscreteVar=NULL){

    if(!is.null(path) && !is.null(data)) stop ("Please do not specify two data sources")
    if (!is.null(data)) setData(data)
    if (!is.null(path)) setData(read.csv(path, sep=",",header=T))

    if(is.null(maxStatesDiscreteVar)) setMaxStatesDiscreteVariable(10)
    else                              setMaxStatesDiscreteVariable(maxStatesDiscreteVar)

  }
)
##################################################################################################

# Gets the data type of a variable.
#
# Args:
#   indexVar: The column index of the data frame where the variable is located.
#
# Returns:
#   A String indicating the type of the variable ("categorical" or "continuous").

Database$methods (
  getDataType = function(indexVar) {
    dataVar = getData()[,indexVar]
    if (!is.numeric(dataVar) ||
        (is.integer(dataVar) && (length(unique(dataVar)) <= getMaxStatesDiscreteVariable()))){
      return("categorical")
    }
    return("continuous")
  }
)

##################################################################################################

# It randomly divides the database into two sets: train and test.
#
# Args:
#   p: Percentage of test records (between 0 and 1).
#   seed: A seed to avoid random divisions (non mandatory argument).
#
# Returns:
#   A List with two objects: train and test.

Database$methods (
  divideIntoTrainAndTest = function(p, seed=NULL) {
    n = length(data[,1])
    tamTest =round(p*n)
    if(!is.null(seed))
      set.seed(seed)
    index = sample(1:n,tamTest)
    datatest = data[sort(index),]
    #datatest = data[index,]
    dbtest = Database$new(data=datatest)
    datatrain = data[-index,]
    dbtrain = Database$new(data=datatrain)
    return(list(dbtrain=dbtrain,dbtest=dbtest))
  }
)

##################################################################################################

# Projects a database over a set of variables.
#
# Args:
#   colIndexes: A vector with the indexes of the variables to be projected.
#
# Returns:
#   The projected database.

Database$methods (
  projectColIndexes = function(colIndexes) {
    return(Database$new(data=data[,colIndexes]))
  }
)

##################################################################################################

# Projects a database over a set of variables.
#
# Args:
#   colIndexes: A vector with the variables to be projected.
#
# Returns:
#   The projected database.

Database$methods (
  projectVariables = function(variables) {
    varNames = c()
    #No consigo hacerlo con lapply
    for(var in variables){
      varNames = c(varNames, var$getName())
    }
    return(Database$new(data=data.frame(data[, varNames])))
  }
)

##################################################################################################

# Selects the cases whose values for a continuous variable are in the interval [a,b].
#
# Args:
#   var: The continuous variable to be considered.
#   a: The lower bound of the interval.
#   b: The upper bound of the interval.
#
# Returns:
#   The database resulting from the selection.

Database$methods (
  selectDataFromVariableInterval = function(var, a, b) {
    if(var$isContinuous()){
      selectedCases <- data[data[,var$getName()]>=a & data[,var$getName()]<b,]

      # Since the interval [a,b) is open on the right side, if b is equal to the maximum
      # value of the variable, we include it in the selected cases, because otherwise
      # it would be discarded.
      if(var$getMax()==b)
        selectedCases <- rbind(selectedCases, data[data[,var$getName()]==b,])

      return(Database$new(data=selectedCases))
    }
  }
)

##################################################################################################

# Count the number of rows in the data.
#
# Returns:
#   The number of cases contained in the database.

Database$methods (
  getNumberOfCases = function() {
    return(nrow(data))
  }
)

##################################################################################################

# Count the number of columns in the data.
#
# Returns:
#   The number of variables contained in the database.

Database$methods (
  getNumberOfVars = function() {
      return(ncol(data))
  }
)

##################################################################################################

# Selects the cases whose values for a categorical variable are equal to an state s.
#
# Args:
#   var: the categorical variable to be considered.
#   s: the state of the categorical variable
#
# Returns:
#   The database resulting from the selection.

Database$methods (
  selectDataFromVariableState = function(var, s) {
    if(var$isCategorical()){
      dataVar=data[, var$getName()]
      selectedCases = subset(data,dataVar == s)
      return(Database$new(data=selectedCases))
    }
  }
)

##################################################################################################

# Shows relevant information about the database: #variables, #cases.
#
# Args:
#
# Returns:
#   No arguments are returned. Only the information is displayed on the screen.

Database$methods (
  show = function() {
    cat("Number of variables: ", getNumberOfVars(), "\n")
    cat("    Number of cases: ", getNumberOfCases(), "\n")
  }
)

##################################################################################################

# Gets the name of the variable located at a position of the database.
#
# Args:
#   indexCol: The index of the variable.
#
# Returns:
#   A String with the name of the variable.

Database$methods (
  getNameVar = function(indexCol){
    colnames(getData())[indexCol]
  }
)

##################################################################################################

# Gets the minimum value for a variable
#
# Args:
#   indexCol: The index of the variable.
#
# Returns:
#   The minimum value.

Database$methods (
  getMin = function(indexCol){
    min(getData()[,indexCol])
  }
)

##################################################################################################

# Gets the maximum value for a variable.
#
# Args:
#   indexCol: The index of the variable.
#
# Returns:
#   The maximum value.

Database$methods (
  getMax = function(indexCol){
    max(getData()[,indexCol])
  }
)

##################################################################################################

# Gets the states for a categorical variable.
#
# Args:
#   indexCol: The index of the variable.
#
# Returns:
#   The states of the categorical variable.

Database$methods (
  getStates = function(indexCol){
    names(table(factor(getData()[,indexCol])))
  }
)

##################################################################################################

# Gets the values of a variable.
#
# Args:
#   var: The variable.
#
# Returns:
#   A Vector of values with the data.

Database$methods (
  getColumnData = function(var){
    projectVariables(list(var))$getData()[,1]
  }
)

##################################################################################################

# Gets the values of a variable.
#
# Args:
#   var: The variable.
#
# Returns:
#   A Vector of values with the data.

Database$methods (
  getColumnDataAt = function(i){
    getData()[,i]
  }
)

##################################################################################################

# Merges the cases of the current database to those contained in the database given as parameter.
# Both databases must be consistent in the variables.
#
# Args:
#   dataset: The dataset.
#
# Returns:
#   The merged database.

Database$methods (
  mergeCases = function(dataset){
    .self$setData(rbind(.self$getData(),dataset$getData()))
  }
)

##################################################################################################

# Suffles the cases of the database.
#
# Args:
#   seed: A seed to avoid random suffles (non mandatory argument).
#
# Returns:
#   The suffled database.

Database$methods (
  suffleData = function(seed=NULL) {
    if(!is.null(seed))
      set.seed(seed)
    n = getNumberOfCases()
    setData(getData()[sample(1:n,n),])
  }
)

##################################################################################################

# Gets the test data corresponding to a i-iteration in a k-fold CV.
#
# Args:
#   k: the number of folds.
#   i: the corresponding iteration.
# Returns:
#   The test data.

Database$methods (
  getTestCV = function(i,k) {
    nCases = getNumberOfCases()
    #cat("\n Total number of cases:", nCases)
    sizeBlock = round(nCases/k)
    #cat("\n Block size:", sizeBlock)
    a = (i-1)*sizeBlock+1
    b = (i)*sizeBlock
    if(i==k) b=nCases
    #cat("\n TEST: cases [",a,"-",b,"]")

    test = Database$new(data=getData()[c(a:b),])
    return(test)
  }
)

##################################################################################################

# Gets the train data corresponding to a i-iteration in a k-fold CV.
#
# Args:
#   k: the number of folds.
#   i: the corresponding iteration.
# Returns:
#   The train data.

Database$methods (
  getTrainCV = function(i,k) {

    nCases = getNumberOfCases()
    #cat("\n Total number of cases:", nCases)
    sizeBlock = round(nCases/k)
    #cat("\n Block size:", sizeBlock)
    a = (i-1)*sizeBlock+1
    b = (i)*sizeBlock
    #cat("\n a=",a)
    #cat("\n b=",b)
    if(i==k) b=nCases
    #cat("\n TRAIN: all cases except [",a,"-",b,"]")

    train = Database$new(data=getData()[-c(a:b),])
    return(train)
  }
)

##################################################################################################

# Gets the test dataset corresponding to a i-iteration in a k-fold cross validation keeping the
# proportion of observations for each of the class values.
#
# It is assumed that the class variable appears last in the data.
#
# Args:
#   k: the number of folds.
#   i: the corresponding iteration.
# Returns:
#   The test dataset.
# Author: jsagrado

Database$methods (
  getTestSet = function(i,k) {
    class_observations <- data[,ncol(data)]
    class_values <- unique(class_observations)
    testObservations <- c()

    for (class_value in class_values) {
      key <- which(class_observations[]==class_value)
      nCases <- length(key)
      if (nCases >= k) {
        blockSize = round(nCases/k)
        a <- (i-1)*blockSize+1
        b <- (i)*blockSize
        if(i == k)
          b <- nCases
      } else {
        if (i == 1) {
          a <- 1
          b <- nCases
        } else {
          a <- 0
          b <- 0
        }
      }
      if (a > 0)
        testObservations <- c(testObservations, key[a:b])
    }

    test = Database$new(data=getData()[testObservations,])
    return(test)
  }
)

##################################################################################################

# Gets the train dataset corresponding to a i-iteration in a k-fold cross validation keeping the
# proportion of observations for each of the class values.
#
# It is assumed that the class variable appears last in the data..
#
# Args:
#   k: the number of folds.
#   i: the corresponding iteration.
# Returns:
#   The train dataset.
# Author: jsagrado

Database$methods (
  getTrainSet = function(i,k) {
    class_observations <- data[,ncol(data)]
    class_values <- unique(class_observations)
    testObservations <- c()

    for (class_value in class_values) {
      key <- which(class_observations[]==class_value)
      nCases <- length(key)
      if (nCases >= k) {
        blockSize = round(nCases/k)
        a <- (i-1)*blockSize+1
        b <- (i)*blockSize
        if(i == k)
          b <- nCases
      } else {
        if (i == 1) {
          a <- 1
          b <- nCases
        } else {
          a <- 0
          b <- 0
        }
      }
      if (a > 0)
        testObservations <- c(testObservations, key[a:b])
    }

    train = Database$new(data=getData()[-testObservations,])
    return(train)
  }
)

##################################################################################################

# Scale the data to the range [0,1]
#
# Args:
#
# Returns:
#   The standarized database.

Database$methods (
  scaleData = function() {
    for(i in 1:getNumberOfVars()){
      if(getDataType(i)=="continuous"){
        a = getMin(i); b= getMax(i)
        data[,i] <<- (data[,i]-a)/(b-a)
      }
    }
  }
)

##################################################################################################

# Standarize the continuous data of the database.
#
# Args:
#
# Returns:
#   The standarized database.

Database$methods (
  standardizeData = function() {
    for(i in 1:getNumberOfVars()){
      if(getDataType(i)=="continuous"){
        mu = mean(getColumnDataAt(i))
        sigma = stats::sd(getColumnDataAt(i))
        data[,i] <<- (data[,i]-mu)/sigma
      }
    }
  }
)

##################################################################################################

# Discretize a vector of continuous data using the Equal Frecuency method.
#
# Args:
#   data: a vector of continuous data.
#   intervals: the number of intervals in the discretization.
#
# Returns:
#   The discretized database.

Database$methods (
  discretizeDataEF = function (data,intervals) {
    n=length(data)
    ciclo=ceiling(n/intervals)
    y=data
    for(i in 1:(intervals-1)) {
      y[order(data)[((i-1)*ciclo+1):(i*ciclo)]]=i
    }
    y[order(data)[((intervals-1)*ciclo+1):n]]=intervals
    return(y)
}
)
##################################################################################################

# Discretize the continuous variables of the dataset using the Equal Frecuency method.
#
# Args:
#   intervals: the number of intervals in the discretization.
#
# Returns:
#   The discretized database.

Database$methods (
  discretizeDatasetEF = function(intervals) {
   # discData = getData()
    for(i in 1:getNumberOfVars()) {
      if(getDataType(i)=="continuous"){
        data[,i] <<- as.character(discretizeDataEF(as.vector(getData()[,i]),intervals)-1)
      }
    }
  }
)

##################################################################################################

# Checks whether the dataset is discretized or not
#
# Args:
#
# Returns:
#   FALSE: if any of the columns contains continuous data
#   TRUE: otherwise

Database$methods (
  isDiscretized = function() {
    for(i in 1:getNumberOfVars()) {
      if(getDataType(i)=="continuous"){
        return (FALSE);
      }
    }
    return (TRUE);
  }
)


#
# #######################################################################################
# scaleData=function(dataset){
#   for(i in 1:ncol(dataset)){
#     dataset[,i] = (dataset[,i]-min(dataset[,i]))/(max(dataset[,i])-min(dataset[,i]))
#   }
#   return(as.data.frame(dataset))
# }


